# (c) 2004 Nuxeo SARL <http://nuxeo.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 2 as published
# by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
# 02111-1307, USA.
#
# $Id: cpscatalog_install.py,v 1.1.2.1 2004/05/20 14:07:56 ben Exp $
"""This is a script to reconfigure portal_catalog indexes for CPS 3.0.x

HOWTO USE THAT ?

 - Log into the ZMI as manager
 - Go to your CPS root directory
* - Create an External Method with the following parameters:

     id    : cpscatalog_installer (or whatever)
     title : CPSCatalog optimizer installer
     Module Name   :  CPSDefault.cpscatalog_install
     Function Name : install

 - save it
 - click now the test tab of this external method.
 - that's it !

"""


import os
import sys
from Products.ZCTextIndex.ZCTextIndex import manage_addLexicon, ZCTextIndex
from Products.PluginIndexes.TopicIndex.TopicIndex import TopicIndex
from zLOG import LOG, INFO

class CPS30xCatalogInstaller:
    """This is a back port of cps 3.1.x catalog optimisation.

    Copying some codes from CMFInstaller and cpsinstaller."""

    DEFAULT_CPS_LEXICON_ID = 'cps_defaut_lexicon'
    DEFAULT_CPS_LEXICON_TITLE = 'CPS Default Lexicon for ZCTextIndex'

    CPS_FILTER_SETS_INDEX = 'cps_filter_sets'
    # this set is supposed to match anything that can be viewed
    # ie anything except document in portal_repository
    CPS_FILTER_SEARCHABLE_SET = 'searchable'
    CPS_FILTER_SEARCHABLE_EXPR = """not filter(lambda s:s.startswith('portal_') or s and s[0] in ('.', '_'), o.getPhysicalPath())"""
    CPS_FILTER_LEAVES_SET = 'leaves'
    CPS_FILTER_LEAVES_EXPR = """o.portal_type not in ('Section', 'Workspace')"""

    def __init__(self, context):
        self.context = context
        self.messages = []
        self.portal = context.portal_url.getPortalObject()

    #
    # Logging
    #
    def log(self, message):
        self.messages.append(message)
        LOG("CPS30xCatlogInstaller", INFO, message)

    def logOK(self):
        pass

    def flush(self):
        log = '\n'.join(self.messages)
        self.messages = []
        return log

    def logResult(self):
        # Wrap HTML around it if it's the main installer.
        return '''<html><head><title>%s</title></head>
            <body><pre>%s</pre></body></html>''' % (
            'CPS30xCatlogInstaller', self.flush() )

    #
    # Catalog stuff
    #
    def flagCatalogForReindex(self, indexid=None):
        if indexid is None:
            # Reindex all catalog
            self.portal._v_reindex_catalog = 1
            return
        indexes = getattr(self.portal, '_v_changed_indexes', [])
        indexes.append(indexid)
        self.portal._v_changed_indexes = indexes

    def reindexCatalog(self):
        # Reindex portal_catalog
        reindex_catalog = getattr(self.portal, '_v_reindex_catalog', 0)
        changed_indexes = getattr(self.portal, '_v_changed_indexes', [])
        if reindex_catalog or len(changed_indexes) > 1:
            self.log('Reindex Catalog')
            self.portal.portal_catalog.refreshCatalog(clear=1)
        elif changed_indexes:
            ct = self.portal.portal_catalog
            self.log('Reindex Catalog indexes %s' % ', '.join(changed_indexes))
            for name in changed_indexes:
                ct.reindexIndex(name, self.portal.REQUEST)
        setattr(self.portal, '_v_reindex_catalog', 0)
        setattr(self.portal, '_v_changed_indexes', [])

    def addZCTextIndexLexicon(self, id, title=''):
        """Add a ZCTextIndex Lexicon."""
        self.log(' Adding a ZCTextIndex Lexicon %s' % id)
        ct = self.portal.portal_catalog
        if id in ct.objectIds():
            self.logOK()
            return

        class Struct:
            def __init__(self, **kw):
                for k, v in kw.items():
                    setattr(self, k, v)
        elems = (Struct(group='Case Normalizer',
                        name='Case Normalizer',
                        ),
                 Struct(group='Stop Words',
                        name=" Don't remove stop words"
                        #name="Remove listed and single char words"
                        #name="Remove listed stop words only"
                        ),
                 Struct(group='Word Splitter',
                        name='Whitespace splitter',
                        ))
        manage_addLexicon(ct, id, title, elems)
        self.log(' Added')

    def addPortalCatalogIndex(self, id, type, extra=None):
        """Adds an index on portal_catalog."""
        self.log(' Portal_catalog verify index %s: %s' % (type, id))
        ct = self.portal.portal_catalog
        if id in ct.indexes():
            current_index_type = ct._catalog.getIndex(id).meta_type
            if current_index_type == type:
                self.logOK()
                return
            elif type == 'ZCTextIndex' and current_index_type != 'TextIndex':
                # we only turn TextIndex into ZCTextIndex keeping NG
                self.log('  keeping index of type %s.' % current_index_type)
                return
            self.log('  Deleting old index')
            ct.delIndex(id)
        self.log('  Adding index')
        if type == 'ZCTextIndex':
            index = ZCTextIndex(id, extra=extra, caller=ct)
            ct._catalog.addIndex(id, index)
        else:
            ct.addIndex(id, type, extra)

        if type == 'TopicIndex' and extra:
            index = ct._catalog.getIndex(id)
            for filter in extra:
                index.addFilteredSet(filter.id,
                                     'PythonFilteredSet',
                                     filter.expr)
                self.log('   Adding filterSet %s for TopicIndex %s' %
                         (filter.id, id))

        self.flagCatalogForReindex(id)

    def addPortalCatalogMetadata(self, id, default_value=None):
        """Adds a metadata in the portal_catalog."""
        self.log(' Portal_catalog verify Metadata: %s, default value: %s'
                 % (id, default_value))
        ct = self.portal.portal_catalog
        if not ct._catalog.schema.has_key(id):
            self.log('  Adding metadata')
            ct.addColumn(id, default_value)

    #
    # Index definition
    #
    def catalogEnumerateIndexes( self ):
        #   Return a list of ( index_name, type ) pairs for the initial
        #   index set.
        class Struct:
            def __init__(self, **kw):
                for k, v in kw.items():
                    setattr(self, k, v)

        return (('Title', 'FieldIndex', None) # used for sorting
                , ('ZCTitle', 'ZCTextIndex',  # used for searching
                   Struct(doc_attr='Title',
                          lexicon_id=self.DEFAULT_CPS_LEXICON_ID,
                          index_type='Okapi BM25 Rank'))
                , ('Subject', 'KeywordIndex', None)
                , ('Description', 'ZCTextIndex',
                   Struct(doc_attr='Description',
                          lexicon_id=self.DEFAULT_CPS_LEXICON_ID,
                          index_type='Okapi BM25 Rank'))
                , ('Creator', 'FieldIndex', None)
                , ('SearchableText', 'ZCTextIndex',
                   Struct(doc_attr='SearchableText',
                          lexicon_id=self.DEFAULT_CPS_LEXICON_ID,
                          index_type='Okapi BM25 Rank'))
                , ('Date', 'DateIndex', None)
                , ('Type', 'FieldIndex', None)
                , ('created', 'DateIndex', None)
                , ('effective', 'DateIndex', None)
                , ('expires', 'DateIndex', None)
                , ('modified', 'DateIndex', None)
                , ('allowedRolesAndUsers', 'KeywordIndex', None)
                , ('review_state', 'FieldIndex', None)
                , ('in_reply_to', 'FieldIndex', None)
                , ('meta_type', 'FieldIndex', None)
                , ('id', 'FieldIndex', None)
                , ('getId', 'FieldIndex', None)
                , ('path', 'PathIndex', None)
                , ('portal_type', 'FieldIndex', None)
                , (self.CPS_FILTER_SETS_INDEX, 'TopicIndex',
                   (Struct(id=self.CPS_FILTER_SEARCHABLE_SET,
                           expr=self.CPS_FILTER_SEARCHABLE_EXPR),
                    Struct(id=self.CPS_FILTER_LEAVES_SET,
                           expr=self.CPS_FILTER_LEAVES_EXPR),))
                , ('start', 'DateIndex', None)
                , ('end', 'DateIndex', None)
                , ('time', 'DateIndex', None) # time of the last transition
               )

    def catalogEnumerateMetadata( self ):
        #   Return a sequence of schema names to be cached (catalog metadata).
        #   id is depricated and may go away, use getId!
        return ('Subject'               # CMF metadata
                , 'Title'
                , 'Description'
                , 'Type'
                , 'review_state'
                , 'Creator'
                , 'Date'
                , 'getIcon'
                , 'created'
                , 'effective'
                , 'expires'
                , 'modified'
                , 'CreationDate'
                , 'EffectiveDate'
                , 'ExpiresDate'
                , 'ModificationDate'
                , 'id'
                , 'getId'
                , 'portal_type'
                # CPS metadata
                , 'Contributors'
                , 'Language'
                , 'start'
                , 'end'
                , 'getRevision'
                , 'time'                # time of the last transition
               )

    def setupCatalog(self):
        # check default ZCTextIndex lexicon
        self.addZCTextIndexLexicon(self.DEFAULT_CPS_LEXICON_ID,
                                   self.DEFAULT_CPS_LEXICON_TITLE)
        ct = self.portal.portal_catalog
        # check indexes
        for index_name, index_type, index_extra in \
                self.catalogEnumerateIndexes():
            self.addPortalCatalogIndex(index_name, index_type, index_extra)
        # check metadata
        for name in self.catalogEnumerateMetadata():
            self.addPortalCatalogMetadata(name)
        # reindex if needed
        self.reindexCatalog()


def install(self):
    installer = CPS30xCatalogInstaller(self)
    installer.log("Starting CPS Catalog optimizer install")
    installer.setupCatalog()
    installer.log("End of CPS Catalog optimizer.")
    return installer.logResult()
